module Unify where
import Start
----------------------------------------------------------------
-- To test this: see `README' or type `make'
----------------------------------------------------------------
type Var = Int
class Children t where   children    :: t -> [t]  
                         mapChildren :: (t->t) -> t -> t
class VarCheck t where   varCheck    :: t -> Maybe Var
class TopEq t    where   topEq       :: t -> t -> Bool

-- Terms have children, variables and top level equality
class (Children t, VarCheck t, TopEq t) => Term t

----------------------------------------------------------------
subTerms :: Children t => t -> [t]
subTerms t = t : concat (map subTerms (children t))

vars   :: (Children t,VarCheck t) => t -> [Var]
vars t =  [ v | Just v <- map varCheck (subTerms t)]

appSubst :: (Subst s, Term t) => s t -> t -> t 
appSubst s t = case varCheck t of
                 Nothing -> mapChildren (appSubst s) t
                 Just v  -> case lookupIn s v of
                              Nothing -> t 
                              Just t' -> appSubst s t'


type Con = String
data T = V' Var| App Con [T]  

instance Children T where children      (V' v)     = []
			  children      (App c ts) = ts
			  mapChildren f (V' v)     = V' v
			  mapChildren f (App c ts) = App c (map f ts)

instance VarCheck T where varCheck      (V' v)      = Just v
			  varCheck      _     = Nothing

instance TopEq T where  topEq (App c l) (App c' l') = c == c'  && 
					    length l == length l' 
			topEq (V' v)    (V' w)      = v == w
			topEq _ _        = False


----------------------------------------------------------------
-- Substitutions 

class Subst s where
  idSubst  :: s t
  modBind  :: (Var,t) -> s t -> s t
  lookupIn :: s t -> Var -> Maybe t

occursCheck :: (Term t,Subst s) => Var -> s t -> t -> Bool
occursCheck i s t = i `elem` reachlist (vars t)
  where 
    reachlist l = l ++ concat (map reachable l)
    reachable v = reachlist (maybe [] vars (lookupIn s v))
        
----------------------------------------------------------------
-- Unification
unify       :: (Term t,Subst s) =>  t  ->  t  ->        Maybe (s t)
unify'      :: (Term t,Subst s) =>  t  ->  t  -> s t -> Maybe (s t)

unify tx ty  =  unify' tx ty idSubst

unify' tx ty =  uni (varCheck tx,varCheck ty) where
  uni (Nothing,Nothing) | topEq tx ty =  uniTerms tx ty
			| otherwise   =  err
  uni (Just i,Just j)   | i == j      =  ok
  uni (Just i, _ )      	      =  i |-> ty
  uni ( _ ,Just j)		      =  j |-> tx

  uniTerms x y = threadList (zipWith unify' (children x) (children y))

(|->) :: (Term t, Subst s) => Var -> t ->  s t -> Maybe (s t)
(i |-> t) s = if occursCheck i s t then  err s
              else case lookupIn s i of 
	             Nothing  ->  ok (modBind (i,t) s)
	             Just t'  ->  mapMaybe (modBind (i,t)) (unify' t t' s)

threadList :: Monad m => [a -> m a] -> a -> m a
threadList = foldr (@@) return

(@@)   :: Monad m => (a -> m b) -> (c -> m a) -> (c -> m b)
(f @@ g) x = g x >>= f

mapMaybe :: (a -> b) -> Maybe a -> Maybe b
mapMaybe f = maybe Nothing (Just . f)
-- name clash with (unused) Maybe.mapMaybe (Haskell Library func.)
{- just map works fine too:
mapMaybe = fmap 
-}
{-
-- defined in the polytypic part
ok, err  :: a -> Maybe a
ok  = Just 
err = const Nothing
-}
----------------------------------------------------------------
-- end of unification
----------------------------------------------------------------
----------------------------------------------------------------
-- instances generated by PolyP are tied in by hand 

----------------------------------------------------------------
-- d a has updatable children
--instance Regular d => Children (d a) where
--  children    = children'
--  mapChildren = mapChildren'
instance Children (VarTree a) where
  children = children'_f7VarTree
  mapChildren = mapChildren'_f7VarTree

instance Children (VarList a) where
  children = children'_f7VarList
  mapChildren = mapChildren'_f7VarList

instance Children (SimpleType a) where
  children = children'_f10SimpleType
  mapChildren = mapChildren'_f10SimpleType

----------------------------------------------------------------
-- d a has top level equality
--instance (Regular d, Eq a) => TopEq (d a) where
--  topEq = topEq'
instance (Eq a) => TopEq (VarTree a) where
  topEq = topEq'_f7VarTree

instance (Eq a) => TopEq (VarList a) where
  topEq = topEq'_f7VarList

instance (Eq a) => TopEq (SimpleType a) where
  topEq = topEq'_f10SimpleType

----------------------------------------------------------------
-- 
instance VarCheck (VarTree a) where
  varCheck = varCheck'_f7VarTree

instance VarCheck (VarList a) where
  varCheck = varCheck'_f7VarList

instance VarCheck (SimpleType a) where
  varCheck = varCheck'_f10SimpleType

----------------------------------------------------------------
-- FuncSubst is a substitution 
data FuncSubst t = FSub (Var -> Maybe t)

instance Subst FuncSubst where
  idSubst = FSub (const Nothing)
  modBind (i,t) (FSub s) = FSub (\j-> if i==j then Just t else s j)
  lookupIn (FSub s) i = s i

----------------------------------------------------------------
-- Alternative definition of appSubst
--  more in the spirit of polytypic programming
{-
appSubst s = mapTerm (mapMaybe (appSubst s) . lookupIn s)

mapTerm :: Term t => (Var -> Maybe t) -> t -> t
mapTerm f t = maybe (mapChildren (mapTerm f) t) 
                    (maybe t id . f) 
                    (varCheck t)
-}
